from sqlalchemy import and_, or_, desc, asc, func, bindparam
from sqlalchemy.orm import subqueryload, contains_eager
from inbox.api.err import InputError
from inbox.api.validation import valid_public_id
from inbox.models import (Contact, Event, Calendar, Message,
                          MessageContactAssociation, Thread,
                          Block, Part, MessageCategory, Category,
                          Metadata)
from inbox.models.event import RecurringEvent
from inbox.sqlalchemy_ext.util import bakery
from inbox.ignition import engine_manager
from inbox.models.session import session_scope_by_shard_id


def contact_subquery(db_session, namespace_id, email_address, field):
    return db_session.query(Message.thread_id) \
        .join(MessageContactAssociation) \
        .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
        .filter(
            Contact.email_address == email_address,
            Contact.namespace_id == namespace_id,
            MessageContactAssociation.field == field) \
        .subquery()


def threads(namespace_id, subject, from_addr, to_addr, cc_addr, bcc_addr,
            any_email, thread_public_id, started_before, started_after,
            last_message_before, last_message_after, filename, in_, unread,
            starred, limit, offset, view, db_session):

    if view == 'count':
        query = db_session.query(func.count(Thread.id))
    elif view == 'ids':
        query = db_session.query(Thread.public_id)
    else:
        query = db_session.query(Thread)

    filters = [Thread.namespace_id == namespace_id, Thread.deleted_at == None]
    if thread_public_id is not None:
        filters.append(Thread.public_id == thread_public_id)

    if started_before is not None:
        filters.append(Thread.subjectdate < started_before)

    if started_after is not None:
        filters.append(Thread.subjectdate > started_after)

    if last_message_before is not None:
        filters.append(Thread.recentdate < last_message_before)

    if last_message_after is not None:
        filters.append(Thread.recentdate > last_message_after)

    if subject is not None:
        filters.append(Thread.subject == subject)

    query = query.filter(*filters)

    if from_addr is not None:
        from_query = contact_subquery(db_session, namespace_id,
                                      from_addr, 'from_addr')
        query = query.filter(Thread.id.in_(from_query))

    if to_addr is not None:
        to_query = contact_subquery(db_session, namespace_id,
                                    to_addr, 'to_addr')
        query = query.filter(Thread.id.in_(to_query))

    if cc_addr is not None:
        cc_query = contact_subquery(db_session, namespace_id,
                                    cc_addr, 'cc_addr')
        query = query.filter(Thread.id.in_(cc_query))

    if bcc_addr is not None:
        bcc_query = contact_subquery(db_session, namespace_id,
                                     bcc_addr, 'bcc_addr')
        query = query.filter(Thread.id.in_(bcc_query))

    if any_email is not None:
        any_contact_query = db_session.query(Message.thread_id) \
            .join(MessageContactAssociation) \
            .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
            .filter(Contact.email_address.in_(any_email),
                    Contact.namespace_id == namespace_id)\
            .subquery()
        query = query.filter(Thread.id.in_(any_contact_query))

    if filename is not None:
        files_query = db_session.query(Message.thread_id). \
            join(Part).join(Block). \
            filter(Block.filename == filename,
                   Block.namespace_id == namespace_id). \
            subquery()
        query = query.filter(Thread.id.in_(files_query))

    if in_ is not None:
        category_filters = [Category.name == in_, Category.display_name == in_]
        try:
            valid_public_id(in_)
            category_filters.append(Category.public_id == in_)
        except InputError:
            pass
        category_query = db_session.query(Message.thread_id). \
            prefix_with('STRAIGHT_JOIN'). \
            join(Message.messagecategories).join(MessageCategory.category). \
            filter(Category.namespace_id == namespace_id,
                   or_(*category_filters)).subquery()
        query = query.filter(Thread.id.in_(category_query))

    if unread is not None:
        read = not unread
        unread_query = db_session.query(Message.thread_id).filter(
            Message.namespace_id == namespace_id,
            Message.is_read == read).subquery()
        query = query.filter(Thread.id.in_(unread_query))

    if starred is not None:
        starred_query = db_session.query(Message.thread_id).filter(
            Message.namespace_id == namespace_id,
            Message.is_starred == starred).subquery()
        query = query.filter(Thread.id.in_(starred_query))

    if view == 'count':
        return {"count": query.one()[0]}

    # Eager-load some objects in order to make constructing API
    # representations faster.
    if view != 'ids':
        expand = (view == 'expanded')
        query = query.options(*Thread.api_loading_options(expand))

    query = query.order_by(desc(Thread.recentdate)).limit(limit)

    if offset:
        query = query.offset(offset)

    if view == 'ids':
        return [x[0] for x in query.all()]

    return query.all()


def messages_or_drafts(namespace_id, drafts, subject, from_addr, to_addr,
                       cc_addr, bcc_addr, any_email, thread_public_id,
                       started_before, started_after, last_message_before,
                       last_message_after, received_before, received_after,
                       filename, in_, unread, starred, limit, offset, view,
                       db_session):
    # Warning: complexities ahead. This function sets up the query that gets
    # results for the /messages API. It loads from several tables, supports a
    # variety of views and filters, and is performance-critical for the API. As
    # such, it is not super simple.
    #
    # We bake the generated query to avoid paying query compilation overhead on
    # every request. This requires some attention: every parameter that can
    # vary between calls *must* be inserted via bindparam(), or else the first
    # value passed will be baked into the query and reused on each request.
    # Subqueries (on contact tables) can't be properly baked, so we have to
    # call query.spoil() on those code paths.

    param_dict = {
        'namespace_id': namespace_id,
        'drafts': drafts,
        'subject': subject,
        'from_addr': from_addr,
        'to_addr': to_addr,
        'cc_addr': cc_addr,
        'bcc_addr': bcc_addr,
        'any_email': any_email,
        'thread_public_id': thread_public_id,
        'received_before': received_before,
        'received_after': received_after,
        'started_before': started_before,
        'started_after': started_after,
        'last_message_before': last_message_before,
        'last_message_after': last_message_after,
        'filename': filename,
        'in_': in_,
        'unread': unread,
        'starred': starred,
        'limit': limit,
        'offset': offset
    }

    if view == 'count':
        query = bakery(lambda s: s.query(func.count(Message.id)))
    elif view == 'ids':
        query = bakery(lambda s: s.query(Message.public_id))
    else:
        query = bakery(lambda s: s.query(Message))

        # Sometimes MySQL doesn't pick the right index. In the case of a
        # regular /messages query, ix_message_ns_id_is_draft_received_date
        # is the best index because we always filter on
        # the namespace_id, is_draft and then order by received_date.
        # For other "exotic" queries, we let the MySQL query planner
        # pick the right index.
        if all(v is None for v in [subject, from_addr, to_addr, cc_addr,
                                   bcc_addr, any_email, thread_public_id,
                                   filename, in_, started_before,
                                   started_after, last_message_before,
                                   last_message_after]):
            query += lambda q: q.with_hint(
                Message,
                'FORCE INDEX (ix_message_ns_id_is_draft_received_date)',
                'mysql')

    query += lambda q: q.join(Thread, Message.thread_id == Thread.id)
    query += lambda q: q.filter(
        Message.namespace_id == bindparam('namespace_id'),
        Message.is_draft == bindparam('drafts'),
        Thread.deleted_at == None)

    if subject is not None:
        query += lambda q: q.filter(Message.subject == bindparam('subject'))

    if unread is not None:
        query += lambda q: q.filter(Message.is_read != bindparam('unread'))

    if starred is not None:
        query += lambda q: q.filter(Message.is_starred == bindparam('starred'))

    if thread_public_id is not None:
        query += lambda q: q.filter(
            Thread.public_id == bindparam('thread_public_id'))

    # TODO: deprecate thread-oriented date filters on message endpoints.
    if started_before is not None:
        query += lambda q: q.filter(
            Thread.subjectdate < bindparam('started_before'),
            Thread.namespace_id == bindparam('namespace_id'))

    if started_after is not None:
        query += lambda q: q.filter(
            Thread.subjectdate > bindparam('started_after'),
            Thread.namespace_id == bindparam('namespace_id'))

    if last_message_before is not None:
        query += lambda q: q.filter(
            Thread.recentdate < bindparam('last_message_before'),
            Thread.namespace_id == bindparam('namespace_id'))

    if last_message_after is not None:
        query += lambda q: q.filter(
            Thread.recentdate > bindparam('last_message_after'),
            Thread.namespace_id == bindparam('namespace_id'))

    if received_before is not None:
        query += lambda q: q.filter(
            Message.received_date <= bindparam('received_before'))

    if received_after is not None:
        query += lambda q: q.filter(
            Message.received_date > bindparam('received_after'))

    if to_addr is not None:
        query.spoil()
        to_query = db_session.query(MessageContactAssociation.message_id) \
            .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
            .filter(
                MessageContactAssociation.field == 'to_addr',
                Contact.email_address == to_addr,
                Contact.namespace_id == bindparam('namespace_id')).subquery()
        query += lambda q: q.filter(Message.id.in_(to_query))

    if from_addr is not None:
        query.spoil()
        from_query = db_session.query(MessageContactAssociation.message_id) \
            .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
            .filter(
                MessageContactAssociation.field == 'from_addr',
                Contact.email_address == from_addr,
                Contact.namespace_id == bindparam('namespace_id')).subquery()
        query += lambda q: q.filter(Message.id.in_(from_query))

    if cc_addr is not None:
        query.spoil()
        cc_query = db_session.query(MessageContactAssociation.message_id) \
            .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
            .filter(
                MessageContactAssociation.field == 'cc_addr',
                Contact.email_address == cc_addr,
                Contact.namespace_id == bindparam('namespace_id')).subquery()
        query += lambda q: q.filter(Message.id.in_(cc_query))

    if bcc_addr is not None:
        query.spoil()
        bcc_query = db_session.query(MessageContactAssociation.message_id) \
            .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
            .filter(
                MessageContactAssociation.field == 'bcc_addr',
                Contact.email_address == bcc_addr,
                Contact.namespace_id == bindparam('namespace_id')).subquery()
        query += lambda q: q.filter(Message.id.in_(bcc_query))

    if any_email is not None:
        query.spoil()
        any_email_query = db_session.query(
            MessageContactAssociation.message_id) \
            .join(Contact, MessageContactAssociation.contact_id == Contact.id)\
            .filter(
                Contact.email_address.in_(any_email),
                Contact.namespace_id == bindparam('namespace_id')) \
            .subquery()
        query += lambda q: q.filter(Message.id.in_(any_email_query))

    if filename is not None:
        query += lambda q: q.join(Part).join(Block). \
            filter(Block.filename == bindparam('filename'),
                   Block.namespace_id == bindparam('namespace_id'))

    if in_ is not None:
        query.spoil()
        category_filters = [Category.name == bindparam('in_'),
                            Category.display_name == bindparam('in_')]
        try:
            valid_public_id(in_)
            category_filters.append(Category.public_id == bindparam('in_id'))
            # Type conversion and bindparams interact poorly -- you can't do
            # e.g.
            # query.filter(or_(Category.name == bindparam('in_'),
            #                  Category.public_id == bindparam('in_')))
            # because the binary conversion defined by Category.public_id will
            # be applied to the bound value prior to its insertion in the
            # query. So we define another bindparam for the public_id:
            param_dict['in_id'] = in_
        except InputError:
            pass
        query += lambda q: q.prefix_with('STRAIGHT_JOIN'). \
            join(Message.messagecategories).join(MessageCategory.category). \
            filter(Category.namespace_id == namespace_id,
                   or_(*category_filters))

    if view == 'count':
        res = query(db_session).params(**param_dict).one()[0]
        return {"count": res}

    query += lambda q: q.order_by(desc(Message.received_date))
    query += lambda q: q.limit(bindparam('limit'))
    if offset:
        query += lambda q: q.offset(bindparam('offset'))

    if view == 'ids':
        res = query(db_session).params(**param_dict).all()
        return [x[0] for x in res]

    # Eager-load related attributes to make constructing API representations
    # faster. Note that we don't use the options defined by
    # Message.api_loading_options() here because we already have a join to the
    # thread table. We should eventually try to simplify this.
    query += lambda q: q.options(
        contains_eager(Message.thread),
        subqueryload(Message.messagecategories).joinedload('category'),
        subqueryload(Message.parts).joinedload(Part.block),
        subqueryload(Message.events))

    prepared = query(db_session).params(**param_dict)
    return prepared.all()


def files(namespace_id, message_public_id, filename, content_type,
          limit, offset, view, db_session):

    if view == 'count':
        query = db_session.query(func.count(Block.id))
    elif view == 'ids':
        query = db_session.query(Block.public_id)
    else:
        query = db_session.query(Block)

    query = query.filter(Block.namespace_id == namespace_id)

    # limit to actual attachments (no content-disposition == not a real
    # attachment)
    query = query.outerjoin(Part)
    query = query.filter(or_(Part.id.is_(None),
                             Part.content_disposition.isnot(None)))

    if content_type is not None:
        query = query.filter(or_(Block._content_type_common == content_type,
                                 Block._content_type_other == content_type))

    if filename is not None:
        query = query.filter(Block.filename == filename)

    # Handle the case of fetching attachments on a particular message.
    if message_public_id is not None:
        query = query.join(Message) \
            .filter(Message.public_id == message_public_id)

    if view == 'count':
        return {"count": query.one()[0]}

    query = query.order_by(asc(Block.id)).distinct().limit(limit)

    if offset:
        query = query.offset(offset)

    if view == 'ids':
        return [x[0] for x in query.all()]
    else:
        return query.all()


def filter_event_query(query, event_cls, namespace_id, event_public_id,
                       calendar_public_id, title, description, location, busy):

    query = query.filter(event_cls.namespace_id == namespace_id).filter(
                event_cls.deleted_at == None)  # noqa

    if event_public_id:
        query = query.filter(event_cls.public_id == event_public_id)

    if calendar_public_id is not None:
        query = query.join(Calendar). \
            filter(Calendar.public_id == calendar_public_id,
                   Calendar.namespace_id == namespace_id)

    if title is not None:
        query = query.filter(event_cls.title.like(u'%{}%'.format(title)))

    if description is not None:
        query = query.filter(event_cls.description.like(u'%{}%'
                                                        .format(description)))

    if location is not None:
        query = query.filter(event_cls.location.like(u'%{}%'.format(location)))

    if busy is not None:
        query = query.filter(event_cls.busy == busy)

    query = query.filter(event_cls.source == 'local')

    return query


def recurring_events(filters, starts_before, starts_after, ends_before,
                     ends_after, db_session, show_cancelled=False):
    # Expands individual recurring events into full instances.
    # If neither starts_before or ends_before is given, the recurring range
    # defaults to now + 1 year (see events/recurring.py)

    recur_query = db_session.query(RecurringEvent)
    recur_query = filter_event_query(recur_query, RecurringEvent, *filters)

    if show_cancelled is False:
        recur_query = recur_query.filter(RecurringEvent.status != 'cancelled')

    before_criteria = []
    if starts_before:
        before_criteria.append(RecurringEvent.start < starts_before)
    if ends_before:
        # start < end, so event start < ends_before
        before_criteria.append(RecurringEvent.start < ends_before)
    recur_query = recur_query.filter(and_(*before_criteria))
    after_criteria = []
    if starts_after:
        after_criteria.append(or_(RecurringEvent.until > starts_after,
                                  RecurringEvent.until == None))  # noqa
    if ends_after:
        after_criteria.append(or_(RecurringEvent.until > ends_after,
                                  RecurringEvent.until == None))  # noqa

    recur_query = recur_query.filter(and_(*after_criteria))

    recur_instances = []

    for r in recur_query:
        # the occurrences check only checks starting timestamps
        if ends_before and not starts_before:
            starts_before = ends_before - r.length
        if ends_after and not starts_after:
            starts_after = ends_after - r.length
        instances = r.all_events(start=starts_after, end=starts_before)
        recur_instances.extend(instances)

    return recur_instances


def events(namespace_id, event_public_id, calendar_public_id, title,
           description, location, busy, starts_before, starts_after,
           ends_before, ends_after, limit, offset, view,
           expand_recurring, show_cancelled, db_session):

    query = db_session.query(Event)

    if not expand_recurring:
        if view == 'count':
            query = db_session.query(func.count(Event.id))
        elif view == 'ids':
            query = db_session.query(Event.public_id)

    filters = [namespace_id, event_public_id, calendar_public_id,
               title, description, location, busy]
    query = filter_event_query(query, Event, *filters)

    event_criteria = []

    if starts_before is not None:
        event_criteria.append(Event.start < starts_before)

    if starts_after is not None:
        event_criteria.append(Event.start > starts_after)

    if ends_before is not None:
        event_criteria.append(Event.end < ends_before)

    if ends_after is not None:
        event_criteria.append(Event.end > ends_after)

    if not show_cancelled:
        if expand_recurring:
            event_criteria.append(Event.status != 'cancelled')
        else:
            # It doesn't make sense to hide cancelled events
            # when we're not expanding recurring events,
            # so don't do it.
            # We still need to show cancelled recurringevents
            # for those users who want to do event expansion themselves.
            event_criteria.append(
                (Event.discriminator == 'recurringeventoverride') |
                ((Event.status != 'cancelled') & (Event.discriminator !=
                                                  'recurringeventoverride')))

    event_predicate = and_(*event_criteria)
    query = query.filter(event_predicate)

    if expand_recurring:
        expanded = recurring_events(filters, starts_before, starts_after,
                                    ends_before, ends_after, db_session,
                                    show_cancelled=show_cancelled)

        # Combine non-recurring events with expanded recurring ones
        all_events = query.filter(Event.discriminator == 'event').all() + \
            expanded

        if view == 'count':
            return {"count": len(all_events)}

        all_events = sorted(all_events, key=lambda e: e.start)
        if limit:
            offset = offset or 0
            all_events = all_events[offset:offset + limit]
    else:
        if view == 'count':
            return {"count": query.one()[0]}
        query = query.order_by(asc(Event.start)).limit(limit)
        if offset:
            query = query.offset(offset)
        # Eager-load some objects in order to make constructing API
        # representations faster.
        all_events = query.all()

    if view == 'ids':
        return [x[0] for x in all_events]
    else:
        return all_events


def messages_for_contact_scores(db_session, namespace_id, starts_after=None):
    query = (db_session.query(
        Message.to_addr, Message.cc_addr, Message.bcc_addr,
        Message.id, Message.received_date.label('date'))
        .join(MessageCategory.message)
        .join(MessageCategory.category)
        .filter(Message.namespace_id == namespace_id)
        .filter(Category.name == 'sent')
        .filter(~Message.is_draft)
        .filter(Category.namespace_id == namespace_id))

    if starts_after:
        query = query.filter(Message.received_date > starts_after)

    return query.all()


def metadata(namespace_id, app_id, view, limit, offset,
             db_session):

    if view == 'count':
        query = db_session.query(func.count(Metadata.id))
    elif view == 'ids':
        query = db_session.query(Metadata.object_public_id)
    else:
        query = db_session.query(Metadata)

    filters = [Metadata.namespace_id == namespace_id,
               Metadata.value.isnot(None)]
    if app_id is not None:
        filters.append(Metadata.app_id == app_id)

    query = query.filter(*filters)
    if view == 'count':
        return {"count": query.scalar()}

    query = query.order_by(desc(Metadata.id)).limit(limit)

    if offset:
        query = query.offset(offset)

    if view == 'ids':
        return [x[0] for x in query.all()]

    return query.all()


def metadata_for_app(app_id, limit, last, query_value, query_type, db_session):
    if app_id is None:
        raise ValueError('Must specify an app_id')

    query = db_session.query(Metadata).filter(Metadata.app_id == app_id)
    if last is not None:
        query = query.filter(Metadata.id > last)

    if query_type is not None:
        if query_type not in METADATA_QUERY_OPERATORS:
            raise ValueError(
                'Invalid query operator for metadata query_type. Must be '
                'one of {}'.format(', '.join(METADATA_QUERY_OPERATORS.keys())))
        operator_filter = METADATA_QUERY_OPERATORS[query_type](query_value)
        query = query.filter(operator_filter)

    query = query.order_by(asc(Metadata.id)).limit(limit)
    return query.all()


def page_over_shards(Model, cursor, limit, get_results=lambda q: q.all()):
    # TODO revisit passing lambda, and cursor format
    cursor = int(cursor)
    start_shard_id = engine_manager.shard_key_for_id(cursor)
    results = []
    remaining_limit = limit
    next_cursor = None
    for shard_id in sorted(engine_manager.engines):
        if shard_id < start_shard_id:
            continue

        if len(results) >= limit:
            break

        with session_scope_by_shard_id(shard_id) as mailsync_session:
            latest_cursor = cursor if shard_id == start_shard_id else None
            query = mailsync_session.query(Model)
            if latest_cursor:
                query = query.filter(Model.id > latest_cursor)
            query = query.order_by(asc(Model.id)).limit(remaining_limit)
            latest_results = get_results(query)

            if latest_results:
                results.extend(latest_results)
                last = latest_results[-1]
                if hasattr(last, 'id'):
                    next_cursor = last.id
                elif 'id' in last:
                    next_cursor = last['id']
                else:
                    raise ValueError('Results returned from get_query must'
                                     'have an id')

                # Handle invalid ids
                cursor_implied_shard = next_cursor >> 48
                if shard_id != 0 and cursor_implied_shard == 0:
                    next_cursor += shard_id << 48

                remaining_limit -= len(latest_results)
    return results, str(next_cursor)

METADATA_QUERY_OPERATORS = {
    '>': lambda v: Metadata.queryable_value > v,
    '>=': lambda v: Metadata.queryable_value >= v,
    '<': lambda v: Metadata.queryable_value < v,
    '<=': lambda v: Metadata.queryable_value <= v,
    '==': lambda v: Metadata.queryable_value == v,
    '!=': lambda v: Metadata.queryable_value != v,
}
